/** @file
 *
 * This file contains functionality required for client-side communication
 * within the fwRPC communication model.
 *
 * Inbound messages are processed at most once.
 *
 * Outbound messages are buffered internally if the target ringbuffer is full
 * until the ringbuffer is read from, after which the client will automatically
 * attempt to write again from the buffer.
 */


#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_exception.ctl"
#uses "fwRPC/fwRPC_RingBuffer.ctl"
#uses "fwRPC/fwRPC_UUID.ctl"
#uses "fwRPC/fwRPC_debug.ctl"


/**
 * Return codes
 * @ingroup Constants
 * @ {
 */
global const int fwRPC_MSGCLIENT_E_SUCCESS     = 0;
global const int fwRPC_MSGCLIENT_E_ERROR       = -1;
global const int fwRPC_MSGCLIENT_E_NOT_RUNNING = -2;
/* @} */

/**
 * Client types
 *
 * @TODO: msgClient should not know about RPC stuff
 *
 * @ingroup Constants
 * @ {
 */
const int fwRPC_MSGCLIENT_TYPE_CLIENT = 0;
const int fwRPC_MSGCLIENT_TYPE_SERVER = 1;
const int fwRPC_MSGCLIENT_TYPE_WORKER = 2;

const int fwRPC_MSGCLIENT_MONITOR_INTERVAL_SEC = 2;
/* @} */


/**
 * Whether or not the client is running
 * Structure:
 *  key = client context handle
 *  value = if client with given ctx is running
 */
private global mapping g_fwRPC_msgClient_isRunning;

/**
 * The client name
 * Structure:
 *  key = client context handle
 *  value = client name with given ctx
 */
private global mapping g_fwRPC_msgClient_name;

/**
 * A mapping of message delegate functions.
 * Structure:
 *  key = client context handle
 *  value = MAPPING:
 *            key   = fwRPC_DISTMSG_TYPE_*
 *            value = STRING: function name
 */
private global mapping g_fwRPC_msgClient_msgDelegates;

/**
 *  Outbound message buffer. If the outbound ringbuffer is full, messages are
 *  enqueued here.
 *
 *  Structure:
 *    key = client context handle
 *    value = DYN_STRING:
 *             [x]: message
 */
private global mapping g_fwRPC_msgClient_msgBuffer;

/**
 * Current tid of the dp connection monitor
 * Structure:
 *  key = client context handle
 *  value = monitor thread id of client with given ctx
 */
private global mapping g_fwRPC_msgClient_monitor_tid;


//------------------------------------------------------------------------------


/**
 * Return whether or not the client module is running
 *
 * @returns true if the client module is running
 *          false otherwise
 */
bool fwRPC_msgClient_isRunning(anytype ctx) {
  if (!mappingHasKey(g_fwRPC_msgClient_isRunning, ctx)) return false;
  return g_fwRPC_msgClient_isRunning[ctx];
}


/**
 * Initialise and run a fwRPC message client with a given name. If no name is
 * specified, one is generated and returned as an output param.
 *
 * Accepts a mapping of fwRPC_DIST_TAG_* to delegate function names.
 *  Delegate functions must have the following signature:
 *
 *    void func(string project, string name, anytype msg);
 *
 *  Where project is the origin project,
 *        name is the origin name
 *        msg is the unserialized message payload
 *
 * typeTag can be any of the following:
 *  fwRPC_DIST_TAG_CLIENT
 *  fwRPC_DIST_TAG_CORE
 *  fwRPC_DIST_TAG_WORKER
 *
 *  The msgDelegates mapping must keep to the following structure:
 *    key   = message type (fwRPC_DISTMSG_TYPE_*)
 *    value = function name of STRING
 *
 * @param local        (IN, bool)   whether or not to initialize in local mode
 * @param typeTag      (IN, string) client type (fwRPC_DIST_TAG_*
 * @param msgDelegates (IN, string) mapping of fwRPC_DISTMSG_TYPE_* to function
 *                                  name
 * @param name         (IN|OUT, mapping) client name
 *
 * @returns 1                         if the context was taken over
 *          fwRPC_MSGCLIENT_E_SUCCESS on success
 *          fwRPC_MSGCLIENT_E_ERROR   otherwise
 *
 * @throws if the parameters are invalid
 */
int fwRPC_msgClient_initialize(
  anytype &ctx,
  bool    local,
  string  typeTag,
  string  &name,
  mapping msgDelegates
) {
  string prefix;

  if (name == "") {
    name = fwRPC_UUID_generateUUID4();
  }

  if (
    typeTag != fwRPC_DIST_TAG_CLIENT &&
    typeTag != fwRPC_DIST_TAG_CORE &&
    typeTag != fwRPC_DIST_TAG_WORKER
  ) {
    fwException_throw(
      fwRPC_EX_MSGCLIENT_INVALID_TYPE, "Client type must be a valid type."
    );
  }

  if (!mappinglen(msgDelegates)) fwException_throw(
    fwRPC_EX_INVALID_PARAM,
    "msgDelegates can not be empty"
  );

  ctx = fwRPC_dist_buildClientName(local, typeTag, name);

  bool wasRunning = fwRPC_msgClient_isRunning(ctx);
  if (wasRunning) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_WARNING,
      "Client context " + ctx + " exists, overwriting!"
    );
  }

  g_fwRPC_msgClient_name[ctx] = ctx;

  g_fwRPC_msgClient_msgBuffer[ctx] = makeDynString();
  g_fwRPC_msgClient_msgDelegates[ctx] = msgDelegates;


  string rbMbxDp = fwRPC_common_makeDpName(
    fwRPC_RBMBX_DPT,
    ctx
  );

  if (!dpExists(rbMbxDp)) dpCreate(rbMbxDp, fwRPC_RBMBX_DPT);

  fwRPC_RingBuffer_init(rbMbxDp + ".in");
  fwRPC_RingBuffer_init(rbMbxDp + ".out");

  g_fwRPC_msgClient_isRunning[ctx] = true;

  int tid = startThread("_fwRPC_msgClient_monitor", ctx, rbMbxDp);
  if (tid == -1) {
    g_fwRPC_msgClient_isRunning[ctx] = false;
    return fwRPC_MSGCLIENT_E_ERROR;
  }

  g_fwRPC_msgClient_monitor_tid[ctx] = tid;

  dpSetWait(rbMbxDp + ".initTrigger", true);
  dpSetWait(rbMbxDp + ".initTrigger", false);

  return wasRunning ? 1 : fwRPC_MSGCLIENT_E_SUCCESS;
}


/**
 * Stop the running client module. Disconnects from ringbuffer messagebox
 * updates.
 *
 * @returns fwRPC_MSGCLIENT_E_SUCCESS     on success
 *          fwRPC_MSGCLIENT_E_NOT_RUNNING if not running
 *          fwRPC_MSGCLIENT_E_ERROR       otherwise
 *
 * @throws nothing
 */
int fwRPC_msgClient_stop(anytype ctx) {
  if (!fwRPC_msgClient_isRunning(ctx)) return fwRPC_MSGCLIENT_E_NOT_RUNNING;

  int err = fwRPC_msgClient_sendRequest(ctx, fwRPC_DISTMSG_TYPE_DISCONNECT);
  if (err < 0) return fwRPC_MSGCLIENT_E_ERROR;

  //set early to abort receipt of msg
  g_fwRPC_msgClient_isRunning[ctx] = false;

  //XXX: we cannot wait for the running thread if it is in another CTRL context,
  //      maybe there is another way to ensure that the msgClient has fully shut
  //      down?  --  msgClient tick thread monitors the _isRunning flag so it
  //      should be ok?

  return fwRPC_MSGCLIENT_E_SUCCESS;
}


/**
 * Send a message to a given callee with specific payload.
 *
 * Messages are processed at the receiver at most once.
 *
 * The payload is automatically serialized.
 *
 * @param type    (IN, string) message type
 * @param sysName (IN, string) target system name
 * @param name    (IN, string) target name
 * @param payload (IN, mixed)  message payload (non serialized)
 *
 * @returns fwRPC_MSGCLIENT_E_SUCCESS     on success
 *          fwRPC_MSGCLIENT_E_NOT_RUNNING if the client is not running
 *          fwRPC_MSGCLIENT_E_ERROR       if the buffer could not be written to
 *
 * @throws nothing
 */
synchronized int fwRPC_msgClient_send(
  anytype ctx,
  string type,
  string sysName,
  string name,
  mixed payload = ""
) {
  if (!fwRPC_msgClient_isRunning(ctx)) return fwRPC_MSGCLIENT_E_NOT_RUNNING;

  string msg = fwRPC_dist_buildMessage(
    type,
    sysName,
    name,
    payload
  );

  if (dynlen(g_fwRPC_msgClient_msgBuffer[ctx]) > 0) {
    dynAppend(g_fwRPC_msgClient_msgBuffer[ctx], msg);
    return fwRPC_MSGCLIENT_E_SUCCESS;
  }

  //ringbufferDp should sorta be.. dist_1:fwRPC_RingBuffer/clientName.out
  string ringbufferDp = getSystemName() +
    fwRPC_common_makeDpName(fwRPC_RBMBX_DPT, ctx) + ".out";

  int err = fwRPC_RingBuffer_write(ringbufferDp, msg);

  if (err == fwRPC_RingBuffer_EFULL) {
    dynAppend(g_fwRPC_msgClient_msgBuffer[ctx], msg);
  } else if (err != fwRPC_RingBuffer_ESUCCESS) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "send: Error writing to buffer: " + ringbufferDp + " (" + err + ")"
    );
    return fwRPC_MSGCLIENT_E_ERROR;
  };

  return fwRPC_MSGCLIENT_E_SUCCESS;
}


/** 
 * Send a request with a given payload.
 * The destination is determined by the broker and routed to the correct callee.
 *
 * Messages are processed at the receiver at most once.
 *
 * The payload is automatically serialized.
 *
 * This function behaves the same as fwRPC_msgClient_msg_send with the
 *  destination set to broker defaults.
 *
 * @param type   (IN, string) message type
 * @param paylad (IN, mixed)  message payload (non serialized)
 *
 * @returns fwRPC_MSGCLIENT_E_SUCCESS     on success
 *          fwRPC_MSGCLIENT_E_NOT_RUNNING if the client is not running
 *          fwRPC_MSGCLIENT_E_ERROR       otherwise
 *
 * @throws nothing
 */
synchronized int fwRPC_msgClient_sendRequest(
  anytype ctx,
  string type,
  mixed payload = ""
) {
  return fwRPC_msgClient_send(
    ctx,
    type,
    fwRPC_DIST_BRKR,
    fwRPC_DIST_BRKR,
    payload
  );
}


/**
 * Retrieve the name of the client instance.
 *
 * @returns client name
 *
 * @throws nothing
 */
string fwRPC_msgClient_getName(anytype ctx) {
  return g_fwRPC_msgClient_name[ctx];
}


//------------------------------------------------------------------------------


/**
 * @internal
 *
 * Runs the dp monitor to listen for disconnect requests. Allows for the msg
 * client to be disconnected from arbitrary context.
 *
 * @param rbMbxDp  (IN, string) message box to monitor
 *
 * @returns nothing
 *
 * @throws nothing
 */
private void _fwRPC_msgClient_monitor(anytype ctx, string rbMbxDp) {
  _fwRPC_msgClient_link(ctx, rbMbxDp);

  while (true) {
    if (!fwRPC_msgClient_isRunning(ctx)) {
      _fwRPC_msgClient_unlink(ctx, rbMbxDp);
      return;
    }

    delay(fwRPC_MSGCLIENT_MONITOR_INTERVAL_SEC);
  }
}


/**
 * @internal
 *
 * Connects to the necessary DPs for receiving incoming messages and
 * notification of buffer drain events.
 *
 * @returns nothing
 *
 * @throws if dpConnect fails
 */
private void _fwRPC_msgClient_link(anytype ctx, string rbMbxDp) {
  int err = dpConnectUserData(
    "_fwRPC_msgClient_msgInCB",
    makeDynString(ctx, rbMbxDp + ".in"), true,
    rbMbxDp + ".in" + fwRPC_RingBuffer_DPE_WRITE
  );

  if (err < 0) {
    fwException_throw(fwRPC_EX_RB_CONNECT,
    "Error connecting msgClient input callback.");
  }

  err = dpConnectUserData(
    "_fwRPC_msgClient_rbDrainCB", ctx, false,
    rbMbxDp + ".out" + fwRPC_RingBuffer_DPE_READ
  );

  if (err < 0) {
    fwException_throw(fwRPC_EX_RB_CONNECT,
    "Error connecting msgClient drain callback.");
  }
}


/**
 * @internal
 *
 * Disconnects from the necessary DPs for receiving incoming messages and
 * notifications of buffer drain events.
 *
 * @returns nothing
 *
 * @throws if dpDisconnect fails
 */
private void _fwRPC_msgClient_unlink(anytype ctx, string rbMbxDp) {
  int err = dpDisconnectUserData(
    "_fwRPC_msgClient_msgInCB",
    makeDynString(ctx, rbMbxDp + ".in"),
    rbMbxDp + ".in" + fwRPC_RingBuffer_DPE_WRITE
  );

  if (err < 0) {
    fwException_throw(fwRPC_EX_RB_DISCONNECT,
    "Error disconnecting msgClient input callback.");
  }

  err = dpDisconnectUserData(
    "_fwRPC_msgClient_rbDrainCB", ctx,
    rbMbxDp + ".out" + fwRPC_RingBuffer_DPE_READ
  );

  if (err < 0) {
    fwException_throw(fwRPC_EX_RB_DISCONNECT,
    "Error disconnecting msgClient drain callback.");
  }
}


/**
 * @internal
 *
 * RB_MBX inbound message handler. Callback to dpConnectUserData.
 * Should be connected to a ringbuffer.write_idx.
 *
 * @TODO: See rbDrainCB todo comment
 *
 * @param ringbufferDp (IN) inbound ringbuffer DP
 * @param dpe          (IN, string) triggering .write_idx DP
 * @param writeidx     (IN, string) write index
 */
synchronized private void _fwRPC_msgClient_msgInCB(
  dyn_string userData,// string ringbufferDp,
  string dpe,
  uint   writeidx
) {
  string ctx = userData[1];
  string ringbufferDp = userData[2];

  if (!fwRPC_msgClient_isRunning(ctx)) return;

  string msg;

  int err = fwRPC_RingBuffer_read(ringbufferDp, msg);
  if (err != fwRPC_RingBuffer_ESUCCESS) return;

  string id;
  string key;
  string sysName;
  string name;
  string buf;

  fwRPC_common_unpackMessage(msg, id, key, sysName, name, buf);

  if (id == "") {
    fwRPC_dbg_log(fwRPC_DBG_WARNING,
      "msgClient received message with no id from: " + sysName + " " + name +
      " of type " + key
    );
    return;
  }

  if (!mappingHasKey(g_fwRPC_msgClient_msgDelegates[ctx], key)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_WARNING,
      "client has no delegate for " + key
    );
    return;
  }

  anytype deserialized;
  deserializeFromXml(buf, deserialized);

  startThread(
    g_fwRPC_msgClient_msgDelegates[ctx][key],
    ctx,
    sysName,
    name,
    deserialized
  );
}


/**
 * @internal
 *
 * Ringbuffer drain event CB. Evokes queue processing.
 *
 * Callback for dpConnect. Function should be connected on ringbuffer read_idx.
 *
 * @TODO: Despite using writeMany, this is callback is executed whenever a value
 * is read, so effectively only one value is written at a time leading to at
 * most the same performance as just calling _write(). This would function
 * better if more than one value is read at once. Alternatively some waiting
 * mechanism could be implemented to wait for a reasonable amount of time until
 * attempting to write again.
 *
 * @param _       (IN, string) ignored
 * @param readidx (IN, string) ringbuffer read index
 *
 * @returns nothing
 *
 * @throws nothing
 */
synchronized private void _fwRPC_msgClient_rbDrainCB(
  string ctx,
  string readidx
) {
  //TODO: don't construct the dp name every time..
  string ringbufferDp = getSystemName() +
    fwRPC_common_makeDpName(fwRPC_RBMBX_DPT, g_fwRPC_msgClient_name[ctx]) +
    ".out";

  int to_insert = dynlen(g_fwRPC_msgClient_msgBuffer[ctx]);
  if (to_insert == 0) return;

  int inserted = 0;
  
  int err = fwRPC_RingBuffer_writeMany(
    ringbufferDp,
    g_fwRPC_msgClient_msgBuffer[ctx],
    to_insert,
    inserted
  );

  if (err != fwRPC_RingBuffer_ESUCCESS && inserted == 0) return;

  for (int i=0; i < inserted; ++i) {
    dynRemove(g_fwRPC_msgClient_msgBuffer[ctx], 1);
  }
}
